from scene import *
import ui
from PIL import Image
import io
import json
from maze_tool import MazeTool
from colors import ColorTool

class TilemapPainter(Scene):
  def __init__(self):
    super().__init__()

    # Set up constants (in actual pixels from spritesheet)
    self.TILE_WIDTH = 20  # Half of previous 40
    self.TILE_HEIGHT = 15  # Half of previous 30

    # Initialize undo history
    self.undo_history = []
    self.MAX_UNDO_STEPS = 50  # Limit memory usage

    # Make grid
    self.GRID_COLS = 51
    self.GRID_ROWS = 80

     # Initialize UI state
    self.show_palette = False
    self.selected_tile = None
    self.mode = 'paint'
    self.maze_tool = None

    self.flip_states = [[{'h': False, 'v': False} for _ in range(self.GRID_COLS)]
                        for _ in range(self.GRID_ROWS)]
    self.current_flip_h = False  
    self.current_flip_v = False  

    # Initialize color map for ink/paper colors
    self.color_map = [[{} for _ in range(self.GRID_COLS)]
                     for _ in range(self.GRID_ROWS)]

    # Initialize hole selection state
    self.hole_start = None  # (row, col) of first point
    self.hole_end = None    # (row, col) of second point
    self.selecting_hole = False

    # Load sprite data to get tile indices
    with open('spriteData.json', 'r') as f:
      sprite_data = json.load(f)
      self.tile_indices = {
          name: row * sprite_data['spritesheetCols'] + col
          for name, (col, row) in sprite_data['tiles'].items()
      }

    # Initialize the tilemap grid
    self.tilemap = [[None for _ in range(self.GRID_COLS)]
                   for _ in range(self.GRID_ROWS)]

    # Add tile mapping
    self.tile_names = {}  # Will store index -> name mapping

  def set_color_mode(self, sender):
    self.mode = 'color'
    self.color_tool.randomize_colors()
    self.update_mode_buttons()
    self.update_color_button()

  def update_color_button(self):
    ink = self.color_tool.current_ink
    paper = self.color_tool.current_paper
    self.color_button.background_color = (paper[0], paper[1], paper[2], 1.0)
    self.color_button.tint_color = (ink[0], ink[1], ink[2], 1.0)

  def set_paint_mode(self, sender):
    self.mode = 'paint'
    self.update_mode_buttons()

  def set_fill_mode(self, sender):
    self.mode = 'fill'
    self.update_mode_buttons()

  def set_hole_mode(self, sender):
    self.mode = 'hole'
    self.update_mode_buttons()
    self.reset_hole_selection()

  def reset_hole_selection(self):
    self.hole_start = None
    self.hole_end = None
    self.selecting_hole = False

  def save_state(self):
    """Save current tilemap state to undo history"""
    # Create a deep copy of the current tilemap
    current_state = [[tile for tile in row] for row in self.tilemap]
    self.undo_history.append(current_state)
    # Keep history within size limit
    if len(self.undo_history) > self.MAX_UNDO_STEPS:
      self.undo_history.pop(0)

  def undo_last_action(self):
    """Restore the previous state from history"""
    # Not yet used, an attempt at undo never worked...
    if self.undo_history:
      previous_state = self.undo_history.pop()
      self.tilemap = previous_state
      # Show undo indicator
      self.show_undo_indicator = True
      self.undo_indicator_time = self.t  # Current time from Scene

  def update_mode_buttons(self):
    modes = ['paint', 'fill', 'hole', 'maze', 'color']
    for mode in modes:
      if hasattr(self, f"{mode}_button"):
        button = getattr(self, f"{mode}_button")
        is_selected = self.mode == mode
        button.background_color = '#007AFF' if is_selected else 'white'
        button.tint_color = 'white' if is_selected else 'black'

  def create_hole(self, start_row, start_col, end_row, end_col):

    self.save_state()

    # Ensure start is top-left and end is bottom-right
    top = min(start_row, end_row)
    bottom = max(start_row, end_row)
    left = min(start_col, end_col)
    right = max(start_col, end_col)

    # Draw diagonal line with SHADOW_EDGE
    shadow_tile = self.tile_indices['SHADOW_EDGE']
    void_tile = self.tile_indices['VOID']
    dark_tile = self.tile_indices['DARK_SHADE']

    # Start from top-left
    current_row = top
    current_col = left

    # Draw diagonal line one step at a time until we hit bottom or right edge
    while current_row <= bottom and current_col <= right:
      self.tilemap[current_row][current_col] = shadow_tile
      current_row += 1
      current_col += 1

    # Fill the rest of the rectangle with VOID or DARK based on position relative to diagonal
    for row in range(top, bottom + 1):
      for col in range(left, right + 1):
        # Skip diagonal line cells
        if self.tilemap[row][col] == shadow_tile:
          continue

        # If we're on the same row or further down than we are to the right
        # from the top-left corner, it's in the void area
        rel_row = row - top
        rel_col = col - left
        if rel_row >= rel_col:
          self.tilemap[row][col] = void_tile
        else:
          self.tilemap[row][col] = dark_tile

  def save_tilemap(self, sender):
    # Create new image for the tilemap
    width = self.GRID_COLS * self.TILE_WIDTH
    height = self.GRID_ROWS * self.TILE_HEIGHT
    with Image.new('RGBA', (width, height), (0, 0, 0, 0)) as img:
      # Load spritesheet with PIL
      with Image.open('spritesheet.png') as spritesheet:
        # Copy each tile and apply colors to the appropriate position
        for row in range(self.GRID_ROWS):
          for col in range(self.GRID_COLS):
            tile = self.tilemap[row][col]
            if tile is not None:
              # Calculate source rectangle
              src_x = (tile % self.tiles_per_row) * self.TILE_WIDTH
              src_y = (tile // self.tiles_per_row) * self.TILE_HEIGHT
              src_box = (
                  src_x, src_y,
                  src_x + self.TILE_WIDTH,
                  src_y + self.TILE_HEIGHT
              )
              # Calculate destination position
              dst_x = col * self.TILE_WIDTH
              dst_y = row * self.TILE_HEIGHT

              # Create colored background if colors are set
              if self.color_map[row][col]:
                # Create a solid color image for the background
                paper_color = self.color_map[row][col].get('paper')
                if paper_color:
                  paper_img = Image.new('RGBA',
                                      (self.TILE_WIDTH, self.TILE_HEIGHT),
                                      tuple(int(c * 255) for c in paper_color) + (255,))
                  img.paste(paper_img, (dst_x, dst_y))

              # Copy and possibly tint the tile
              tile_img = spritesheet.crop(src_box)
              if self.color_map[row][col] and self.color_map[row][col].get('ink'):
                # Convert the tile to RGBA if it isn't already
                if tile_img.mode != 'RGBA':
                  tile_img = tile_img.convert('RGBA')
                # Create a solid color image for tinting
                ink_color = self.color_map[row][col]['ink']
                tint_img = Image.new('RGBA', tile_img.size,
                                   tuple(int(c * 255) for c in ink_color) + (255,))
                # Multiply the alpha channels
                tile_img.putalpha(
                    Image.blend(Image.new('L', tile_img.size, 0),
                              tile_img.split()[3] if len(tile_img.split()) > 3 else Image.new('L', tile_img.size, 255),
                              1))
                # Composite the images
                tile_img = Image.composite(tint_img, tile_img, tile_img.split()[3])

              img.paste(tile_img, (dst_x, dst_y), tile_img if tile_img.mode == 'RGBA' else None)

      # Save the image
      img.save('tilemap.png')

  def setup(self):
    try:
      # Load spritesheet dimensions
      with Image.open('spritesheet.png') as img:
        self.sprite_width = img.size[0]/2
        self.sprite_height = img.size[1]/2

      # Calculate tiles based on image size
      self.tiles_per_row = int(self.sprite_width // self.TILE_WIDTH)
      self.tiles_per_col = int(self.sprite_height // self.TILE_HEIGHT)

      # Load tile mappings
      with open('spriteData.json', 'r') as f:
        data = json.load(f)
        for name, coords in data['tiles'].items():
          index = coords[1] * self.tiles_per_row + coords[0]
          self.tile_names[index] = name

    except Exception as e:
      print(f"Error in setup: {e}")
      self.tiles_per_row = 1
      self.tiles_per_col = 1

    self.maze_tool = MazeTool(self.tile_indices)
    self.color_tool = ColorTool(self.tile_indices)

    # Constants for button layout
    BUTTON_WIDTH = 70
    BUTTON_HEIGHT = 30
    BUTTON_SPACING = 10
    TOP_MARGIN = 10

    # First row - tool buttons
    x = BUTTON_SPACING
    y = TOP_MARGIN

    # Create main tool buttons in first row
    buttons = [
        ('💾 Save', self.save_tilemap),
        ('🖌️ Paint', self.set_paint_mode),
        ('🪣 Fill', self.set_fill_mode),
        ('🕳️ Hole', self.set_hole_mode),
        ('🌀 Maze', self.set_maze_mode)
    ]
    
    # Add flip buttons after maze button
    x += BUTTON_WIDTH + BUTTON_SPACING
    flip_h_button = ui.Button(title='↔️')
    flip_h_button.frame = (x, y, BUTTON_HEIGHT, BUTTON_HEIGHT)  # Square button
    flip_h_button.background_color = 'white'
    flip_h_button.tint_color = 'black'
    flip_h_button.corner_radius = 5
    flip_h_button.action = self.toggle_flip_h
    self.flip_h_button = flip_h_button
    self.view.add_subview(flip_h_button)
  
    x += BUTTON_HEIGHT + BUTTON_SPACING
    flip_v_button = ui.Button(title='↕️')
    flip_v_button.frame = (x, y, BUTTON_HEIGHT, BUTTON_HEIGHT)  # Square button
    flip_v_button.background_color = 'white'
    flip_v_button.tint_color = 'black'
    flip_v_button.corner_radius = 5
    flip_v_button.action = self.toggle_flip_v
    self.flip_v_button = flip_v_button
    self.view.add_subview(flip_v_button)

    for title, action in buttons:
      button = ui.Button(title=title)
      button.frame = (x, y, BUTTON_WIDTH, BUTTON_HEIGHT)
      button.background_color = 'white'
      button.tint_color = 'black'
      button.corner_radius = 5
      button.action = action
      self.view.add_subview(button)
      setattr(self, f"{title.split()[1].lower()}_button", button)
      x += BUTTON_WIDTH + BUTTON_SPACING

    # Second row - color palette buttons
    x = BUTTON_SPACING
    y += BUTTON_HEIGHT + BUTTON_SPACING

    # Create exactly 5 color buttons
    self.color_buttons = []
    for i in range(3):  # Create exactly 5 buttons to match color pairs
      color_button = ui.Button(title='▲')
      color_button.frame = (x, y, BUTTON_WIDTH, BUTTON_HEIGHT)
      color_button.corner_radius = 3
      color_button.tag = i
      color_button.action = self.select_color_pair
      color_button.border_width = 1
      color_button.border_color = 'gray'
      self.view.add_subview(color_button)
      self.color_buttons.append(color_button)
      x += BUTTON_WIDTH + BUTTON_SPACING

    # Add refresh colors button at end of second row
    refresh_button = ui.Button(title='🔄')
    refresh_button.frame = (x, y, BUTTON_WIDTH, BUTTON_HEIGHT)
    refresh_button.background_color = 'white'
    refresh_button.tint_color = 'black'
    refresh_button.corner_radius = 5
    refresh_button.action = self.refresh_colors
    self.view.add_subview(refresh_button)

    # Add selected tile label below buttons
    tile_label = ui.Label()
    tile_label.frame = (BUTTON_SPACING, y + BUTTON_HEIGHT + BUTTON_SPACING,
                      self.view.width - 2 * BUTTON_SPACING, BUTTON_HEIGHT)
    tile_label.text = 'No tile selected'
    tile_label.text_color = 'black'
    tile_label.alignment = ui.ALIGN_LEFT
    self.tile_label = ui.Label()
    self.tile_label.frame = (BUTTON_SPACING, y + BUTTON_HEIGHT + BUTTON_SPACING,
                      self.view.width - 2 * BUTTON_SPACING, BUTTON_HEIGHT)
    self.tile_label.text = 'No tile selected'  # Initial text
    self.tile_label.text_color = 'black'
    self.tile_label.alignment = ui.ALIGN_LEFT
    self.view.add_subview(self.tile_label)

  def update_tile_label(self, text):
    """Update the tile label with new text"""
    if hasattr(self, 'tile_label'):
      self.tile_label.text = text

  def select_color_pair(self, sender):
    """Handle color pair button selection"""
    self.mode = 'color'
    self.color_tool.set_current_pair(sender.tag)
    self.update_mode_buttons()
    self.update_color_buttons()
    # Update the label to indicate color mode
    self.update_tile_label(f'Color Mode: {sender.tag + 1}/5')

  def refresh_colors(self, sender):
    """Generate new color pairs"""
    self.color_tool.regenerate_colors()
    self.update_color_buttons()

  def update_color_buttons(self):
    """Update the appearance of color palette buttons"""
    print("Updating color buttons")
    print(f"Number of buttons: {len(self.color_buttons)}")
    print(f"Number of color pairs: {len(self.color_tool.color_pairs)}")

    # Make sure we have the same number of buttons as color pairs
    while len(self.color_buttons) > len(self.color_tool.color_pairs):
      button = self.color_buttons.pop()
      button.remove_from_superview()

    for i, button in enumerate(self.color_buttons):
      if i < len(self.color_tool.color_pairs):
        colors = self.color_tool.color_pairs[i]
        print(f"Color pair {i}:", colors)

        # Convert RGB floats to format suitable for ui.Button
        paper_color = tuple(colors['paper']) + (1.0,)
        ink_color = tuple(colors['ink']) + (1.0,)

        # Apply colors directly to button
        button.background_color = paper_color
        button.tint_color = ink_color
        button.title = '▲'  # Triangle symbol for ink color

      # Highlight selected pair
      button.border_width = 2 if i == self.color_tool.current_pair_index else 1
      button.border_color = 'black' if i == self.color_tool.current_pair_index else 'gray'

  def flood_fill(self, row, col, target_tile, replacement_tile):
    """Iterative flood fill using a stack. Handles None (empty) tiles."""
    # Early return if trying to replace with same tile (including None)
    if target_tile == replacement_tile:
      return

    # Validate initial position
    if (row < 0 or row >= self.GRID_ROWS or
        col < 0 or col >= self.GRID_COLS):
      return

    # Stack of cells to process: (row, col)
    stack = [(row, col)]
    visited = set()  # Keep track of visited cells

    while stack:
      current_row, current_col = stack.pop()

      # Skip if already visited
      if (current_row, current_col) in visited:
        continue

      # Check if cell is valid and matches target (including None)
      if (current_row < 0 or current_row >= self.GRID_ROWS or
          current_col < 0 or current_col >= self.GRID_COLS or
          not (self.tilemap[current_row][current_col] == target_tile or
               (self.tilemap[current_row][current_col] is None and target_tile is None))):
        continue

      # Fill current cell and mark as visited
      self.tilemap[current_row][current_col] = replacement_tile
      visited.add((current_row, current_col))

      # Add adjacent cells to stack (4-way connectivity)
      directions = [
          (current_row + 1, current_col),  # down
          (current_row - 1, current_col),  # up
          (current_row, current_col + 1),  # right
          (current_row, current_col - 1)   # left
      ]

      for next_row, next_col in directions:
        if (next_row, next_col) not in visited:
          stack.append((next_row, next_col))
          
  def toggle_flip_h(self, sender):
    """Toggle horizontal flip state for new tiles"""
    self.current_flip_h = not self.current_flip_h
    sender.background_color = '#007AFF' if self.current_flip_h else 'white'
    sender.tint_color = 'white' if self.current_flip_h else 'black'

  def toggle_flip_v(self, sender):
    """Toggle vertical flip state for new tiles"""
    self.current_flip_v = not self.current_flip_v
    sender.background_color = '#007AFF' if self.current_flip_v else 'white'
    sender.tint_color = 'white' if self.current_flip_v else 'black'
    
  def place_tile(self, row, col):
    """Place a tile based on current mode"""
    if not (0 <= row < self.GRID_ROWS and 0 <= col < self.GRID_COLS):
      return

    if self.mode == 'maze':
      self.save_state()
      self.tilemap = self.maze_tool.fill(self.tilemap, row, col,
                                       self.GRID_ROWS, self.GRID_COLS)
    elif self.selected_tile is not None:
      self.save_state()  # Save state before any tile modification
      if self.mode == 'paint':
        self.tilemap[row][col] = self.selected_tile
      elif self.mode == 'fill':
        target_tile = self.tilemap[row][col]  # This can be None for empty tiles
        self.flood_fill(row, col, target_tile, self.selected_tile)
  
  def draw_tile(self, tile_index, x, y, row=None, col=None):
    """Draw a single tile from the spritesheet with optional color and flipping"""
    if tile_index is None:
      return
  
    # Draw paper color if available
    if row is not None and col is not None and self.color_map[row][col]:
      paper_color = self.color_map[row][col].get('paper')
      if paper_color:
        fill(*paper_color, 1)
        rect(x, y, self.TILE_WIDTH, self.TILE_HEIGHT)
  
    # Calculate source coordinates in actual image pixels
    src_x = (tile_index % self.tiles_per_row) * self.TILE_WIDTH
    src_y = self.sprite_height - ((tile_index // self.tiles_per_row + 1) * self.TILE_HEIGHT)
  
    # Apply ink color if available
    if row is not None and col is not None and self.color_map[row][col]:
      ink_color = self.color_map[row][col].get('ink')
      if ink_color:
        tint(*ink_color, 1)
      else:
        tint(1, 1, 1, 1)
    else:
      tint(1, 1, 1, 1)
  
    # Get flip state for the tile
    flip_h = False
    flip_v = False
    if row is not None and col is not None:
      flip_state = self.flip_states[row][col]
      flip_h = flip_state.get('h', False)  # Use get() with default False
      flip_v = flip_state.get('v', False)  # Use get() with default False
  
    # Calculate actual width and height for flipping
    width = -self.TILE_WIDTH if flip_h else self.TILE_WIDTH
    height = -self.TILE_HEIGHT if flip_v else self.TILE_HEIGHT
  
    # Adjust x and y position for flipping
    dest_x = x + (self.TILE_WIDTH if flip_h else 0)
    dest_y = y + (self.TILE_HEIGHT if flip_v else 0)
  
    # Draw the tile with flipping
    image('spritesheet.png',
          dest_x, dest_y,
          width, height,
          src_x, src_y,
          self.TILE_WIDTH, self.TILE_HEIGHT)
        
  def set_maze_mode(self, sender):
    self.mode = 'maze'
    self.update_mode_buttons()

  def draw(self):
    # Clear background
    background(0.8, 0.8, 0.8)

    # First pass: Draw grid lines only in empty cells
    stroke(0.7, 0.7, 0.7, 0.3)
    stroke_weight(1)
    
    # Draw vertical grid lines
    for col in range(self.GRID_COLS + 1):
        x = col * self.TILE_WIDTH
        # Check adjacent cells for content
        for row in range(self.GRID_ROWS):
            y = (self.GRID_ROWS - 1 - row) * self.TILE_HEIGHT
            
            # Check if both adjacent cells are empty
            left_empty = col == 0 or (
                col > 0 and 
                self.tilemap[row][col-1] is None and 
                not self.color_map[row][col-1]
            )
            right_empty = col == self.GRID_COLS or (
                col < self.GRID_COLS and 
                self.tilemap[row][col] is None and 
                not self.color_map[row][col]
            )
            
            # Only draw grid line segment if at least one adjacent cell is empty
            if left_empty or right_empty:
                line(x, y, x, y + self.TILE_HEIGHT)

    # Draw horizontal grid lines
    for row in range(self.GRID_ROWS + 1):
        y = (self.GRID_ROWS - row) * self.TILE_HEIGHT
        # Check adjacent cells for content
        for col in range(self.GRID_COLS):
            x = col * self.TILE_WIDTH
            
            # Check if both adjacent cells are empty
            top_empty = row == 0 or (
                row > 0 and 
                self.tilemap[row-1][col] is None and 
                not self.color_map[row-1][col]
            )
            bottom_empty = row == self.GRID_ROWS or (
                row < self.GRID_ROWS and 
                self.tilemap[row][col] is None and 
                not self.color_map[row][col]
            )
            
            # Only draw grid line segment if at least one adjacent cell is empty
            if top_empty or bottom_empty:
                line(x, y, x + self.TILE_WIDTH, y)

    # Second pass: Draw colored backgrounds and tiles
    for row in range(self.GRID_ROWS):
        for col in range(self.GRID_COLS):
            x = col * self.TILE_WIDTH
            y = (self.GRID_ROWS - 1 - row) * self.TILE_HEIGHT

            # Draw paper color if available
            if self.color_map[row][col]:
                paper_color = self.color_map[row][col].get('paper')
                if paper_color:
                    fill(*paper_color, 1)
                    no_stroke()
                    rect(x, y, self.TILE_WIDTH, self.TILE_HEIGHT)

            # Draw tile if present
            tile = self.tilemap[row][col]
            if tile is not None:
                self.draw_tile(tile, x, y, row, col)

    # Draw palette on top if active
    if self.show_palette:
        fill(1, 1, 1, 1)
        rect(0, 0, self.size.w, self.size.h)
        fill(0, 0, 0, 0.2)
        rect(0, 0, self.size.w, self.size.h)

        total_tiles = self.tiles_per_row * self.tiles_per_col
        for i in range(total_tiles):
            row = i // self.tiles_per_row
            col = i % self.tiles_per_row
            x = col * self.TILE_WIDTH
            y = (self.tiles_per_col - 1 - row) * self.TILE_HEIGHT
            self.draw_tile(i, x, y)

  def touch_began(self, touch):
    if touch.location.x < self.GRID_COLS * self.TILE_WIDTH and \
       touch.location.y < self.GRID_ROWS * self.TILE_HEIGHT:
      if not self.show_palette:
        col = int(touch.location.x // self.TILE_WIDTH)
        row = self.GRID_ROWS - 1 - int(touch.location.y // self.TILE_HEIGHT)

        if self.mode == 'color':
          self.color_tool.apply_colors(self.tilemap, self.color_map, row, col)
        elif self.mode == 'hole':
          if not self.selecting_hole:
            self.hole_start = (row, col)
            self.selecting_hole = True
          else:
            self.hole_end = (row, col)
            self.create_hole(self.hole_start[0], self.hole_start[1],
                           self.hole_end[0], self.hole_end[1])
            self.reset_hole_selection()
        else:  # For paint, fill, and maze modes
          self.place_tile(row, col)
    else:
      self.show_palette = not self.show_palette

  def touch_moved(self, touch):
    if not self.show_palette:
      col = int(touch.location.x // self.TILE_WIDTH)
      row = self.GRID_ROWS - 1 - int(touch.location.y // self.TILE_HEIGHT)
      if 0 <= col < self.GRID_COLS and 0 <= row < self.GRID_ROWS:
        if self.mode == 'paint' and self.selected_tile is not None:
          self.place_tile(row, col)
        elif self.mode == 'color':
          self.color_tool.apply_colors(self.tilemap, self.color_map, row, col)

  def touch_ended(self, touch):
    if self.show_palette:
      # Select tile from palette
      col = int(touch.location.x // self.TILE_WIDTH)
      row = self.tiles_per_col - 1 - int(touch.location.y // self.TILE_HEIGHT)
      tile_index = row * self.tiles_per_row + col
      if tile_index < self.tiles_per_row * self.tiles_per_col:
        self.selected_tile = tile_index
        # Update the label with tile name if available
        tile_name = self.tile_names.get(tile_index, f'Tile {tile_index}')
        self.update_tile_label(f'Selected: {tile_name}')
        self.show_palette = False

  def touch_moved(self, touch):
    if not self.show_palette:
      col = int(touch.location.x // self.TILE_WIDTH)
      row = self.GRID_ROWS - 1 - int(touch.location.y // self.TILE_HEIGHT)
      if 0 <= col < self.GRID_COLS and 0 <= row < self.GRID_ROWS:
        if self.mode == 'paint' and self.selected_tile is not None:
          self.place_tile(row, col)
        elif self.mode == 'color':
          self.color_tool.apply_colors(self.tilemap, self.color_map, row, col)


if __name__ == '__main__':
  run(TilemapPainter(), PORTRAIT)

